#include "game.h"

#include "camera.h"
#include "resources.h"
#include "object.h"
#include "timer.h"
#include "font.h"
#include "map.h"
#include "player.h"

#define CAMERA_ACCEL 10.f
#define EDITOR_SPEED 600.f
#define EDITOR_ACCEL 10.f

#define CAMERA_RADIUS 80.f
#define CAMERA_DRAG 5.f

Game::Game() {
	//width = 2560;
	//height = 1600;
	//width = 1260;
	//height = 768;
	width = 1024;
	height = 768;
	fullscreen = false;
	compatibilityMode = true;

	editLevel= curTile = curItem = 0;
	hideInactive = true;
	editItems = false;

	cameraModeNb = 2;
	cameraMode = 0;
	cameraOffset = Vector(0, 400, 200);

	resources = new Resources("data/");
	testMesh = new Mesh("data/ld48.msh");

	display = NULL;
	help = false;

	timer = new Timer();
}

Game::~Game() {
	delete resources;
	delete timer;
	delete defaultCamera;
}

bool Game::init() {

	// SDL Init
	SDL_Init(SDL_INIT_VIDEO);
	SDL_SetVideoMode(width, height, 0, SDL_OPENGL | (fullscreen ? SDL_FULLSCREEN : 0));
	SDL_WM_SetCaption("TheTower48 by Codexus", NULL);
	SDL_ShowCursor(SDL_DISABLE);

	// GLEW Init
	glewInit();

	if (GLEW_ARB_vertex_program && GLEW_ARB_fragment_program) {
		compatibilityMode = false;
	}

	// OpenGL Init
	glViewport(0, 0, width, height);
	glClearColor(0, 0, 0, 1);

	glCullFace(GL_BACK);
	glFrontFace(GL_CCW);
	glEnable(GL_CULL_FACE);

	glClearDepth(1.0f);
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);

	// init DevIL
	ilInit();
	ilutRenderer(ILUT_OPENGL);

	// Game init
	defaultCamera = new LookAtCamera();
	defaultCamera->aspectRatio = float(width) / height;
	defaultCamera->pos = Vector(200, 4000, 0);
	defaultCamera->rot = Vector(-80, 0, 0);

	editorCamera = new Camera();
	editorCamera->aspectRatio = float(width) / height;
	editorCamera->pos = Vector(200, 500, 0);
	editorCamera->rot = Vector(-80, 0, 0);

	//bgColor = Vector(.18, .18, .26);
	bgColor = Vector(0, 0, 0);

	world = new World("data/world.txt");
	player = new Player();

	scene = new Object();

	scene->children.push_back(world);
	scene->children.push_back(player);

	startPosition = Vector(0,3800,0);
	lastSavePoint = startPosition;
	player->pos = startPosition;
	curTarget = startPosition;

	lastLevelReached = 0;

	mode = GM_INIT;

	return true;
}

int Game::start() {
	init();

	done = false;
	SDL_Event event;
	while(!done) {
		while(SDL_PollEvent(&event)) {
			switch (event.type) {
				case SDL_KEYDOWN:
					if (mode != GM_EDITOR) {
						if (event.key.keysym.sym == SDLK_ESCAPE) {
							done = true;
						} else if (mode == GM_GAMEOVER && event.key.keysym.sym == SDLK_SPACE) {
							restartGame(false, false);
						} else if (mode != GM_EDITOR && event.key.keysym.sym == SDLK_e && (event.key.keysym.mod & KMOD_CTRL)) {
							restartGame(true, true);
							SDL_ShowCursor(SDL_ENABLE);
							editorCamera->pos.y = (editLevel+2) * FLOOR_SIZE - 10.f;
						} else if (mode == GM_NORMAL && event.key.keysym.sym == SDLK_p) {
							mode = GM_PAUSE;
						} else if (mode == GM_PAUSE && event.key.keysym.sym == SDLK_p) {
							mode = GM_NORMAL;
						} else if (event.key.keysym.sym == SDLK_c) {
							cameraMode = modulo(cameraMode+1, cameraModeNb);
						} else if (event.key.keysym.sym == SDLK_j && (event.key.keysym.mod & KMOD_CTRL)  && (event.key.keysym.mod & KMOD_SHIFT)) {
							player->feetOnGround = false;
							player->speed = Vector(0, 4000, 0);
						} else if (event.key.keysym.sym == SDLK_F1) {
							help = !help;
						}

					} else {
						if (event.key.keysym.sym == SDLK_ESCAPE || (event.key.keysym.sym == SDLK_e && (event.key.keysym.mod & KMOD_CTRL))) {
							SDL_ShowCursor(SDL_DISABLE);
							restartGame(true, false);
						} else if (event.key.keysym.sym == SDLK_PAGEUP) {
							editLevel = modulo(editLevel + 1, int(world->maps.size()));
							editorCamera->pos.y = (editLevel+2) * FLOOR_SIZE - 10.f;
						} else if (event.key.keysym.sym == SDLK_PAGEDOWN) {
							editLevel = modulo(editLevel - 1, int(world->maps.size()));
							editorCamera->pos.y = (editLevel+2) * FLOOR_SIZE - 10.f;
						} else if (event.key.keysym.sym == SDLK_INSERT) {
							curTile = modulo(curTile + 1, int(world->tiles.size()+1));
						} else if (event.key.keysym.sym == SDLK_DELETE) {
							curTile = modulo(curTile - 1, int(world->tiles.size()+1));
						} else if (event.key.keysym.sym == SDLK_HOME) {
							curItem = modulo(curItem + 1, int(world->items.size()+1));
						} else if (event.key.keysym.sym == SDLK_END) {
							curItem = modulo(curItem - 1, int(world->items.size()+1));
						} else if (event.key.keysym.sym == SDLK_RETURN) {
							hideInactive = !hideInactive;
						} else if (event.key.keysym.sym == SDLK_SPACE) {
							editItems = !editItems;
						} else if (event.key.keysym.sym == SDLK_s && (event.key.keysym.mod & KMOD_CTRL)) {
							world->save();
						}
					}
					break;
				case SDL_QUIT:
					done = false;
					break;
			}
		}

		done |= loop();
	}

	SDL_Quit();
	return 0;
}

bool Game::loop() {
	float delta = timer->tick();
	delta = min(.2, delta);

	//editor stuff
	Vector mousePoint;
	int mapx, mapy;

	//
	// Updates

	if (mode == GM_EDITOR) {
		int mx, my;
		int mouseButtons = SDL_GetMouseState(&mx, &my);

		editorCamera->use();
		Vector p1 = editorCamera->pos;
		Vector p2 = editorCamera->getMouseVector(mx, my);
		mousePoint = planeRayIntersection(p1, p1 + p2, Vector(0, 1, 0), -editLevel * FLOOR_SIZE);

		Map *map = world->maps[editLevel];
		map->worldToMap(mousePoint.x, mousePoint.z, &mapx, &mapy);

		map->editorPreview(mapx, mapy, editItems ? 1 : 0, editItems ? curItem : curTile);

		if (mouseButtons & SDL_BUTTON(1)) {
			map->set(mapx, mapy, editItems ? 1 : 0, editItems ? curItem : curTile);
		} else if (mouseButtons & SDL_BUTTON(3)) {
			if (editItems) {
				curItem = map->get(mapx, mapy, 1);
			} else {
				curTile = map->get(mapx, mapy, 0);
			}	
		}  else if (mouseButtons & SDL_BUTTON(2)) {
			map->set(mapx, mapy, editItems ? 1 : 0, 0);
		}

		Uint8 *keys = SDL_GetKeyState(NULL);
		if (keys[SDLK_UP]) {
			editorCamSpeed.z = smoothChange(editorCamSpeed.z, -EDITOR_SPEED, EDITOR_ACCEL, delta);
		} else if (keys[SDLK_DOWN]) {
			editorCamSpeed.z = smoothChange(editorCamSpeed.z, EDITOR_SPEED, EDITOR_ACCEL, delta);
		} else {
			editorCamSpeed.z = smoothChange(editorCamSpeed.z, 0, EDITOR_ACCEL, delta);
		}

		if (keys[SDLK_LEFT]) {
			editorCamSpeed.x = smoothChange(editorCamSpeed.x, -EDITOR_SPEED, EDITOR_ACCEL, delta);
		} else if (keys[SDLK_RIGHT]) {
			editorCamSpeed.x = smoothChange(editorCamSpeed.x, EDITOR_SPEED, EDITOR_ACCEL, delta);
		} else {
			editorCamSpeed.x = smoothChange(editorCamSpeed.x, 0, EDITOR_ACCEL, delta);
		}

		if (keys[SDLK_q]) {
			editorCamSpeed.y = smoothChange(editorCamSpeed.y, -EDITOR_SPEED, EDITOR_ACCEL, delta);
		} else if (keys[SDLK_a]) {
			editorCamSpeed.y = smoothChange(editorCamSpeed.y, EDITOR_SPEED, EDITOR_ACCEL, delta);
		} else {
			editorCamSpeed.y = smoothChange(editorCamSpeed.y, 0, EDITOR_ACCEL, delta);
		}

		editorCamera->pos += editorCamSpeed * delta;
	} else if (mode != GM_PAUSE) {
		UpdateContext uc;
		uc.game = this;
		uc.keys = SDL_GetKeyState(NULL);
		uc.delta = delta;
		uc.gameMode = mode;

		scene->update(&uc);

		// Update camera 
		Vector playerOrientation;
		playerOrientation.x = -sinf(deg2rad(player->rot.y));
		playerOrientation.z = -cosf(deg2rad(player->rot.y));

		if (cameraMode == 1) {
			Vector cameraTargetPos;

			if (uc.keys[SDLK_HOME]) {
				cameraTargetPos = player->pos + Vector(0, 400, 200);
			} else {
				cameraTargetPos = player->pos + Vector(0, 400, 0) + playerOrientation * -200;
			}
			cameraTargetPos.y = max(-200, cameraTargetPos.y);

			Vector cameraVector = cameraTargetPos - defaultCamera->pos;

			float cameraAccel;
			if (uc.keys[SDLK_END] || uc.keys[SDLK_HOME]) {
				cameraAccel = max(0.f, cameraVector.length()) * CAMERA_ACCEL;
			} else {
				cameraAccel = max(0.f, cameraVector.length() - CAMERA_RADIUS) * CAMERA_ACCEL;
			}


			if (cameraAccel > .0001f) {
				cameraSpeed += normal(cameraVector) * cameraAccel * delta;
			}
			defaultCamera->pos += cameraSpeed * delta;

			float newCameraSpeedX = cameraSpeed.x * (1 - CAMERA_DRAG * delta);
			cameraSpeed.x = (newCameraSpeedX * cameraSpeed.x) > 0 ? newCameraSpeedX : 0.f;

			float newCameraSpeedY = cameraSpeed.y * (1 - CAMERA_DRAG * delta);
			cameraSpeed.y = (newCameraSpeedY * cameraSpeed.y) > 0 ? newCameraSpeedY : 0.f;

			float newCameraSpeedZ = cameraSpeed.z * (1 - CAMERA_DRAG * delta);
			cameraSpeed.z = (newCameraSpeedZ * cameraSpeed.z) > 0 ? newCameraSpeedZ : 0.f;
		} else {
			if (uc.keys[SDLK_END]) {
				Vector cameraTargetOffset = playerOrientation * -200.f;
				cameraTargetOffset.y = 400.;

				cameraOffset.x = smoothChange(cameraOffset.x, cameraTargetOffset.x, CAMERA_ACCEL, delta);
				cameraOffset.y = smoothChange(cameraOffset.y, cameraTargetOffset.y, CAMERA_ACCEL, delta);
				cameraOffset.z = smoothChange(cameraOffset.z, cameraTargetOffset.z, CAMERA_ACCEL, delta);
			} else { //if (uc.keys[SDLK_HOME]) {
				Vector cameraTargetOffset = Vector(0, 400, 200);

				cameraOffset.x = smoothChange(cameraOffset.x, cameraTargetOffset.x, CAMERA_ACCEL, delta);
				cameraOffset.y = smoothChange(cameraOffset.y, cameraTargetOffset.y, CAMERA_ACCEL, delta);
				cameraOffset.z = smoothChange(cameraOffset.z, cameraTargetOffset.z, CAMERA_ACCEL, delta);
			}


			Vector cameraTargetPos = player->pos + cameraOffset;
			cameraTargetPos.y = max (-200, cameraTargetPos.y);

			defaultCamera->pos.x = smoothChange(defaultCamera->pos.x, cameraTargetPos.x, CAMERA_ACCEL, delta);
			defaultCamera->pos.y = smoothChange(defaultCamera->pos.y, cameraTargetPos.y, CAMERA_ACCEL, delta);
			defaultCamera->pos.z = smoothChange(defaultCamera->pos.z, cameraTargetPos.z, CAMERA_ACCEL, delta);
		}

		defaultCamera->target.x = smoothChange(defaultCamera->target.x, player->pos.x, CAMERA_ACCEL, delta);
		defaultCamera->target.y = smoothChange(defaultCamera->target.y, player->pos.y, CAMERA_ACCEL, delta);
		defaultCamera->target.z = smoothChange(defaultCamera->target.z, player->pos.z, CAMERA_ACCEL, delta);

		if (player->pos.y < -5000.f) {
			mode = GM_GAMEOVER;
		}
	}

	//
	// Render

	RenderContext rc;
	rc.game = this;

	glClearColor(bgColor.x, bgColor.y, bgColor.z, 1);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	if (mode == GM_EDITOR) {
		editorCamera->use();
	} else {
		defaultCamera->use();
	}

	if (compatibilityMode) {
		startStandardLighting();
		glColor3f(1,1,1);
	} else {
		glBindProgramARB(GL_VERTEX_PROGRAM_ARB, resources->getShader("vtxLight"));

		if (mode == GM_EDITOR) {
			glProgramLocalParameter4fARB(GL_VERTEX_PROGRAM_ARB, 0, editorCamera->pos.x, editorCamera->pos.y, editorCamera->pos.z, 1.f);
			glProgramLocalParameter4fARB(GL_VERTEX_PROGRAM_ARB, 1, editorCamera->pos.x - 500.f, editorCamera->pos.y, editorCamera->pos.z - 200.f, 1.f);
			glProgramLocalParameter4fARB(GL_VERTEX_PROGRAM_ARB, 2, mousePoint.x, mousePoint.y + 20.f, mousePoint.z, 1.f);
		} else {
			glProgramLocalParameter4fARB(GL_VERTEX_PROGRAM_ARB, 0, -500, defaultCamera->pos.y, 300, 1.f);
			glProgramLocalParameter4fARB(GL_VERTEX_PROGRAM_ARB, 1, 500, defaultCamera->pos.y, 200, 1.f);
			glProgramLocalParameter4fARB(GL_VERTEX_PROGRAM_ARB, 2, player->pos.x + 100.f, player->pos.y + 200.f, player->pos.z + 100.f, 1.f);
		}

		GLfloat mat[16];
		glGetFloatv(GL_MODELVIEW_MATRIX, mat);
		glMatrixMode(GL_MATRIX0_ARB);
		glLoadMatrixf(mat);
		glMatrixMode(GL_MODELVIEW);

		//glProgramLocalParameter4fvARB(GL_VERTEX_PROGRAM_ARB, 3, &mat[0]);
		//glProgramLocalParameter4fvARB(GL_VERTEX_PROGRAM_ARB, 4, &mat[4]);
		//glProgramLocalParameter4fvARB(GL_VERTEX_PROGRAM_ARB, 5, &mat[8]);
		//glProgramLocalParameter4fvARB(GL_VERTEX_PROGRAM_ARB, 6, &mat[12]);


		glEnable(GL_VERTEX_PROGRAM_ARB);

		glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB, resources->getShader("frgLight"));

		// 3 lights colors
		if (mode == GM_EDITOR) {
			glProgramLocalParameter4fARB(GL_FRAGMENT_PROGRAM_ARB, 0, 1, 1, 1.f, 300.f);
			glProgramLocalParameter4fARB(GL_FRAGMENT_PROGRAM_ARB, 1, 1, 1, .5f, 100.f);
			glProgramLocalParameter4fARB(GL_FRAGMENT_PROGRAM_ARB, 2, 1, .8, .8, 100.f);
		} else {
			glProgramLocalParameter4fARB(GL_FRAGMENT_PROGRAM_ARB, 0, .5f, .5f, 1.f, 500.f);
			glProgramLocalParameter4fARB(GL_FRAGMENT_PROGRAM_ARB, 1, .4f, .1f, .05f, 300.f);
			glProgramLocalParameter4fARB(GL_FRAGMENT_PROGRAM_ARB, 2, 1.f, 1.f, .8f, 100.f);
		}

		// material specularity
		glProgramLocalParameter4fARB(GL_FRAGMENT_PROGRAM_ARB, 3, 1, 1, 1, 16.f);

		// fog
		//glProgramLocalParameter4fARB(GL_FRAGMENT_PROGRAM_ARB, 4, bgColor.x, bgColor.y, bgColor.z, 1.f);
		//glProgramLocalParameter4fARB(GL_FRAGMENT_PROGRAM_ARB, 5, .997, 10, 1, 1);
		//glProgramLocalParameter4fARB(GL_FRAGMENT_PROGRAM_ARB, 5, .995, 300, 1, 1);

		glEnable(GL_FRAGMENT_PROGRAM_ARB);

		glColor3f(0, 0, 0);
	}

	glBindTexture(GL_TEXTURE_2D, resources->getTexture("test"));
	glEnable(GL_TEXTURE_2D);

	//world->rot.y += delta * 3.f;
	scene->render(&rc);


	if (compatibilityMode) {
		stopStandardLighting();
	} else {
		glDisable(GL_FRAGMENT_PROGRAM_ARB);
		glDisable(GL_VERTEX_PROGRAM_ARB);
	}
	glDisable(GL_TEXTURE_2D);

	// HUD

	glLoadIdentity();
	glScalef(.01, .01, .01);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);

	glDisable(GL_DEPTH_TEST);


	if (mode == GM_EDITOR) {
		glColor3f(1,1,1);
		printText(-100, 95, "EDITOR MODE");
		printText(-100, 90, "Editing level %d (PgUp/PgDn)", editLevel);
		printText(-100, 85, "Current tile %d (Insert/Del)", curTile);
		printText(-100, 80, "Current item %d (Home/End)", curItem);
		printText(-100, 75, "Hide inactive levels (Return): %s", hideInactive ? "on" : "off");
		printText(-100, 70, "Edit items (Space): %s", editItems ? "on" : "off");
		printText(-100, 65, "ESC to exit / Ctrl-S to save");
		printText(-100, 60, "Right click is the tile picker");
		printText(-100, 55, "(%d %d)", mapx, mapy);
	}


	if (mode == GM_NORMAL && help) {
		glColor3f(1,1,1);
		printText(-80, 95, "HELP");

		printText(-80, 90, "Keys:");
		printText(-80, 85, "Arrow keys: move");
		printText(-80, 80, "SPACE: jump");

		printText(-80, 70, "P: pause");
		printText(-80, 65, "C: Switch camera mode");
		printText(-80, 60, "END: View from behind");
		printText(-80, 55, "HOME: Reset camera");

		printText(-80, 45, "CTRL-E: Editor (be careful!!)");

		printText(-80, 35, "ESC: Quit");
	}


	if (mode == GM_PAUSE) {
		glColor3f(1,1,1);
		printText(-20, 0, "GAME PAUSED");
	}


	if (mode == GM_DYING || mode == GM_GAMEOVER) {
		glColor3f(1,0,0);
		printText(-25, 0, "YOU'RE DEAD!");
	}

	if (mode == GM_GAMEOVER) {
		glColor3f(1,0,0);
		printText(-30, -20, "PRESS SPACE TO START AGAIN");
	}

	if (mode == GM_NORMAL) {
		if (display) {
			glColor3f(1,1,1);
			printText(-float(strlen(display)), -0, display);
		}
	}

	if (mode == GM_INIT) {
		glColor3f(1,1,1);
		printText(-10, -0, "GET READY!");
	}

	glColor3f(.5,.5,1);
	printText(-100, -100, "%.1f fps, LD48 rocks! Floor: %d", timer->getFps(), world->getFloorNb(player->pos.y));

	glEnable(GL_DEPTH_TEST);

	display = NULL;

	SDL_GL_SwapBuffers();
	return false;
}


void Game::startStandardLighting() {
	glEnable(GL_LIGHTING);
	glLightfv(GL_LIGHT0, GL_AMBIENT, Vector(.1f, .1f, .1f));
	glLightfv(GL_LIGHT0, GL_DIFFUSE, Vector(.8f, .8f, 1));
	glLightf(GL_LIGHT0, GL_QUADRATIC_ATTENUATION, .000001f);
	glEnable(GL_LIGHT0);

	glLightfv(GL_LIGHT1, GL_AMBIENT, Vector(0, 0, 0));
	glLightfv(GL_LIGHT1, GL_DIFFUSE, Vector(1.f, .8f, .8f));
	glLightf(GL_LIGHT1, GL_QUADRATIC_ATTENUATION, .000002f);
	glEnable(GL_LIGHT1);

	GLfloat pos1[4] = {-500, defaultCamera->pos.y, 300, 1.f};
	GLfloat pos2[4] = {500, defaultCamera->pos.y, 200, 1};

	glLightfv(GL_LIGHT0, GL_POSITION, pos1);
	glLightfv(GL_LIGHT1, GL_POSITION, pos2);
}

void Game::stopStandardLighting() {
	glDisable(GL_LIGHTING);
}

void Game::restartGame(bool levelOne, bool editorMode) {
	mode = editorMode ? GM_EDITOR : GM_INIT;

	// reset player
	player->pos = levelOne ? startPosition : lastSavePoint;
	player->speed = Vector();
	player->feetOnGround = false;
	player->jumpCount = 0;
	player->wantJump = false;
	player->controlX = player->controlY = 0;

	world->resetWorld(levelOne ? 0 : lastLevelReached);

	cameraOffset = Vector(0, 400, 200);

	if (levelOne) {
		lastSavePoint = startPosition;
		lastLevelReached = 0;
	}

}

//int error = glGetError();
//if (error != GL_NO_ERROR) {
//	printf("%s\n", (char *) gluErrorString(error));
//}

	//testMesh->rot.x += delta * 30.f;
	//testMesh->pos.z += delta * 100.f;
	//testMesh->scale.y = 2.f;
	//testMesh->transform();
	//testMesh->render(NULL); //FIXME:
