#include "map.h"
#include "game.h"
#include "resources.h"
#include "player.h"

#define SECTION_SIZE 600

//Map::Map(int width, int height, int depth) {
//	this->width = width;
//	this->height = height;
//	this->depth = depth;
//
//	this->tiles = NULL;
//
//	data = new int[width * height * depth];
//}

Map::Map(string fileName, int level, World *world) {
	this->level = level;
	this->tiles = &world->tiles;
	this->tileSize = Vector(100, 100, 0);
	this->items = &world->items;
	this->world = world;

	this->fileName = fileName;
	load();
}

void Map::load() {
	ifstream file(fileName.c_str());
	file >> width >> height >> depth;
	data = new int[width * height * depth];
	for (int i=0; i < width * height * depth; i++) {
		file >> data[i];
	}
	file.close();

	if (gGame->mode != Game::GM_EDITOR) {
		for (int i = 0; i < width; i++) {
			for (int j = 0; j < height; j++) {
				int item = data[i + j * width + width*height];
				switch (item) {
					case ITEM_ENEMY:
						{
							Enemy *enemy = new Enemy();
							enemy->pos = mapToWorld(i, j);
							enemy->pos.y = world->getFloorPos(level);

							createdObjects.push_back(enemy);
							world->children.push_back(enemy);
							data[i + j * width + width*height] = 0;
						}
						break;
				}
			}
		}
	}
}

void Map::save() {
	ofstream file(fileName.c_str());

	if (file) {
		file << width << " " << height << " " << depth << "\n";
		for (int d = 0; d < depth; d++) {
			for (int j = 0; j < height; j++) {
				for (int i = 0; i < width; i++) {
					file << data[i + j * width + width*height * d] << " ";
				}
				file << "\n";
			}
			file << "\n";
		}
		file.close();
	}
}


Map::~Map() {
	delete[] data;
}

int Map::get(int x, int y, int layer) {
	x = max(0, min(width-1, x));
	y = max(0, min(height-1, y));
	layer = max(0, min(depth-1, layer));

	return data[x + y * width + layer * width * height];
}

void Map::set(int x, int y, int layer, int value) {
	if (x >= 0 && x < width && y >= 0 && y < height && layer >=0 && layer < depth) {
		data[x + y * width + layer * width * height] = value;
	}
}

void Map::render(RenderContext *rc) {
	int tilesNb = (int) tiles->size();
	for (int i = 0; i < width; i++) {
		for (int j = 0; j < height; j++) {
			int tile = data[i + j * width];

			if (gGame->mode == Game::GM_EDITOR && previewL == 0 && previewX == i && previewY == j) {
				tile = previewVal;
			}

			if (tile > 0 && tile <= tilesNb) {
				glPushMatrix();
				glTranslatef((i - width / 2.f) * tileSize.x, 0, (j - height / 2.f) * tileSize.y);
				(*tiles)[tile-1]->render(rc);
				glPopMatrix();
			}
		}
	}

	int itemsNb = (int) tiles->size();
	for (int i = 0; i < width; i++) {
		for (int j = 0; j < height; j++) {
			int tile = data[i + j * width + width*height];

			if (gGame->mode == Game::GM_EDITOR && previewL == 1 && previewX == i && previewY == j) {
				tile = previewVal;
			}

			if (tile > 0 && tile <= itemsNb) {
				glPushMatrix();
				glTranslatef((i - width / 2.f) * tileSize.x, 0, (j - height / 2.f) * tileSize.y);
				(*items)[tile-1]->render(rc);
				glPopMatrix();
			}
		}
	}
	renderChildren(rc);
}

bool Map::onGround(float x, float y, float radius) {
	int i, j;

	for (int c=0; c<16; c++) {
		float offX = radius * cosf(c * M_PI / 8.f);
		float offY = radius * sinf(c * M_PI / 8.f);

		worldToMap(x + offX, y + offY, &i, &j);
		if (i >= 0 && i < width && j >= 0 && j < height) {
			if (get(i, j, 0) > 0) {
				return true;
			}
		}
	}
	
	return false;
}

int Map::getTile(float x, float y, float radius) {
	int i, j;

	for (int c=-1; c<16; c++) {
		float offX = c<0 ? 0.f : (radius * cosf(c * M_PI / 8.f));
		float offY = c<0 ? 0.f : (radius * sinf(c * M_PI / 8.f));

		worldToMap(x + offX, y + offY, &i, &j);
		if (i >= 0 && i < width && j >= 0 && j < height) {
			int tile = get(i, j, 0);
			if (tile > 0) {
				return tile;
			}
		}
	}
	
	return 0;
}

int Map::getItem(float x, float y) {
	int i, j;

	worldToMap(x, y , &i, &j);
	if (i >= 0 && i < width && j >= 0 && j < height) {
		int tile = get(i, j, 1);
		if (tile > 0) {
			return tile;
		}
	}
	
	return 0;
}

void Map::worldToMap(float x, float y, int *i, int *j) {
	*i = (x + (width + 1) * tileSize.x / 2.f) / tileSize.x;
	*j = (y + (height + 1) * tileSize.y / 2.f) / tileSize.y;
}

Vector Map::mapToWorld(int i, int j) {
	return Vector((i - width / 2.f) * tileSize.x, 0, (j - height / 2.f) * tileSize.y);
}

void Map::reset() {
	for(list<Object *>::iterator i = createdObjects.begin(); i != createdObjects.end(); i++) {
		world->children.remove(*i);
		delete *i;
	}
	createdObjects.clear();

	delete[] data;
	load();
}

void Map::editorPreview(int previewX, int previewY, int previewL, int previewVal) {
	this->previewX = previewX;
	this->previewY = previewY;
	this->previewL = previewL;
	this->previewVal = previewVal;
}


//////////////////////////////////////////////////

World::World(string fileName) {
	int levelNb;

	ifstream file(fileName.c_str());
	file >> levelNb;
	file.close();

	tiles.push_back(new Mesh("data/basetile.msh"));
	tiles.push_back(new Mesh("data/icetile.msh"));
	tiles.push_back(new Mesh("data/uptile.msh"));
	tiles.push_back(new Mesh("data/downtile.msh"));
	tiles.push_back(new Mesh("data/lefttile.msh"));
	tiles.push_back(new Mesh("data/righttile.msh"));
	tiles.push_back(new Mesh("data/magtile.msh"));
	tiles.push_back(new Mesh("data/savetile.msh"));

	items.push_back(new Mesh("data/jump.msh"));
	items.push_back(new Mesh("data/evilguy.msh"));
	items.push_back(new Mesh("data/savep.msh"));

	towerSection = new Mesh("data/towersection.msh");

	char buffer[256];
	for (int i = 0; i < levelNb; i++) {
		sprintf(buffer, "data/level%03d.map", i);
		Map *map = new Map(buffer, i, this);
		maps.push_back(map);
	}
}

World::~World() {
}

void World::render(RenderContext *rc) {
	for (int i = -10; i < 9; i++) {
		glPushMatrix();
		glTranslatef(0, i*SECTION_SIZE, 0);
		towerSection->render(rc);
		glPopMatrix();
	}

	if (gGame->mode == Game::GM_EDITOR && gGame->hideInactive) {
		int i = gGame->editLevel;
		glPushMatrix();
		glTranslatef(0, i*FLOOR_SIZE, 0);
		maps[i]->render(rc);
		glPopMatrix();
	} else {
		for (int i = 0; i < int(maps.size()); i++) {
			glPushMatrix();
			glTranslatef(0, i*FLOOR_SIZE, 0);
			maps[i]->render(rc);
			glPopMatrix();
		}
		renderChildren(rc);
	}

}

int World::getFloorNb(float height) {
	return height < 0.f ? int(height / FLOOR_SIZE) - 1 : int(height / FLOOR_SIZE);

}

float World::getFloorPos(int nb) {
	return nb * FLOOR_SIZE;
}

Map *World::getMap(int nb) {
	if (nb >= 0 && nb < int(maps.size())) {
		return maps[nb];
	} else {
		return NULL;
	}
}


void World::resetWorld(int startLevel) {
	for (int i = startLevel; i < int(maps.size()); i++) {
		maps[i]->reset();
	}
}


void World::save() {
	for (int i = 0; i < int(maps.size()); i++) {
		maps[i]->save();
	}
}
