#include "player.h"
#include "game.h"
#include "map.h"
#include "camera.h"

#define PLAYER_SPEED 300.f
#define ENEMY_SPEED 320.f


#define AUTO_SPEED 600.f

#define MINIMUM_SPEED 5.f
#define PLAYER_ACCEL 10.f
#define PLAYER_ACCEL_ROT 10.f

#define GRAVITY -3000.f
#define TERMINAL_VELOCITY -3000.f

#define PLAYER_RADIUS 43.f // used for collisions
#define AI_RADIUS 20.f
#define DEADLY_RADIUS 42.f

#define MAX_JUMP 1
#define JUMP_SPEED 1200.f
#define JUMP_SPEED_2 600.f
#define MEGA_JUMP 2000.f

Player::Player() : Guy("data/blobbyguy.msh") {
	maxSpeed = PLAYER_SPEED;
	type = CREATURE_PLAYER;
}

Player::~Player() {
}

void Player::update(UpdateContext *uc) {
	if (uc->gameMode == Game::GM_INIT && feetOnGround) {
		gGame->mode = Game::GM_NORMAL;
	}

	if (uc->gameMode == Game::GM_NORMAL) {
		int tileType=0, item = 0;
		int curFloor = uc->game->world->getFloorNb(pos.y);
		Map *map = NULL;

		if (feetOnGround) {
			map = uc->game->world->getMap(curFloor); 

			if (map) {
				tileType = map->getTile(pos.x, pos.z, PLAYER_RADIUS);
				item = map->getItem(pos.x, pos.z);
			}
		}

		wantJump = !toggleJump && uc->keys[SDLK_SPACE];
		toggleJump = uc->keys[SDLK_SPACE] != 0;

		Vector control;

		if (uc->keys[SDLK_UP]) {
			control.z = -1.f;
		} else if (uc->keys[SDLK_DOWN]) {
			control.z = 1.f;
		} else {
			control.z = 0;
		}

		if (uc->keys[SDLK_LEFT]) {
			control.x = -1.f;
		} else if (uc->keys[SDLK_RIGHT]) {
			control.x = 1.f;
		} else {
			control.x = 0;
		}

		if (gGame->cameraMode == 1) {
			Vector realControl;
			Vector directionZ = gGame->defaultCamera->pos - gGame->player->pos;
			directionZ.y = 0;
			directionZ.normalize();

			Vector directionX = cross(directionZ, Vector(0, 1, 0));

			realControl = directionZ * control.z;
			realControl += directionX * control.x;

			controlX = min(1.f, max(-1.f, realControl.x));
			controlY = min(1.f, max(-1.f, realControl.z));
		} else {
			controlX = min(1.f, max(-1.f, control.x));
			controlY = min(1.f, max(-1.f, control.z));
		}

		if (tileType == TILE_SAVE) {
			gGame->display = "SAVE POINT - IF YOU DIE YOU'LL RESTART HERE";
		} else if (item == ITEM_JUMP) {
			gGame->display = "JUMP BOOSTER - PRESS SPACE TO JUMP TO NEXT LEVEL";
		}
		
		if (item == ITEM_SAVE) {
			gGame->lastLevelReached = curFloor;
			gGame->lastSavePoint = pos + Vector(0, 100, 0);
			if (map) {
				int i, j;
				map->worldToMap(pos.x, pos.z, &i, &j);
				map->set(i, j, 1, 0);
			}
		}
	}

	Guy::update(uc);
}


///////////////////////////////////////

Guy::Guy(string fileName) : Mesh(fileName) {
	feetOnGround = false;
	jumpCount = 0;
	wantJump = false;
	controlX = controlY = 0.f;
}

Guy::~Guy() {
}


void Guy::update(UpdateContext *uc) {
	int tileType=0, item = 0;
	if (feetOnGround) {
		int curFloor = uc->game->world->getFloorNb(pos.y);
		Map *map = uc->game->world->getMap(curFloor); 

		if (map) {
			tileType = map->getTile(pos.x, pos.z, PLAYER_RADIUS);
			item = map->getItem(pos.x, pos.z);
		}
	}

	float playerSpeed = (tileType == TILE_SLOW) ? maxSpeed / 2.f : maxSpeed;
	Vector targetSpeed;
	targetSpeed.x = playerSpeed * controlX;
	targetSpeed.z = playerSpeed * controlY;

	if (tileType == TILE_LEFT) {
		speed.x = smoothChange(speed.x, -AUTO_SPEED, PLAYER_ACCEL, uc->delta);
		speed.z = smoothChange(speed.z, 0, PLAYER_ACCEL, uc->delta);
	} else if (tileType == TILE_RIGHT) {
		speed.x = smoothChange(speed.x, AUTO_SPEED, PLAYER_ACCEL, uc->delta);
		speed.z = smoothChange(speed.z, 0, PLAYER_ACCEL, uc->delta);
	} else if (tileType == TILE_UP) {
		speed.x = smoothChange(speed.x, 0, PLAYER_ACCEL, uc->delta);
		speed.z = smoothChange(speed.z, -AUTO_SPEED, PLAYER_ACCEL, uc->delta);
	} else if (tileType == TILE_DOWN) {
		speed.x = smoothChange(speed.x, 0, PLAYER_ACCEL, uc->delta);
		speed.z = smoothChange(speed.z, AUTO_SPEED, PLAYER_ACCEL, uc->delta);
	} else if (tileType != TILE_ICE) {
		speed.x = smoothChange(speed.x, targetSpeed.x, PLAYER_ACCEL, uc->delta);
		speed.z = smoothChange(speed.z, targetSpeed.z, PLAYER_ACCEL, uc->delta);
	}

	// gravity
	if (!feetOnGround) {
		speed.y += GRAVITY * uc->delta;
		speed.y = max(TERMINAL_VELOCITY, speed.y);
	} else {
		speed.y = max(0.f, speed.y);
	}

	if (wantJump && jumpCount < MAX_JUMP && tileType != TILE_SLOW && speed.y >= 0) {
		jumpCount++;
		if (item == ITEM_JUMP) {
			speed.y += MEGA_JUMP;
		} else {
			speed.y += (jumpCount == 1) ? JUMP_SPEED : JUMP_SPEED_2;
		}
		feetOnGround = false;
	}


	Vector newPos = pos + speed * uc->delta;
	// Collision detection
	if (!(gGame->mode == Game::GM_DYING && type == CREATURE_PLAYER)) {
		if (feetOnGround) {
			int curFloor = uc->game->world->getFloorNb(newPos.y);
			Map *map = uc->game->world->getMap(curFloor); 
			if (!map || !map->onGround(newPos.x, newPos.z, PLAYER_RADIUS)) {
				feetOnGround = false;
			}
		} else {
			int curFloor = uc->game->world->getFloorNb(pos.y);
			int newFloor = uc->game->world->getFloorNb(newPos.y);
			if (newFloor < curFloor) {
				Map *map = uc->game->world->getMap(curFloor); 
				if (map && map->onGround(newPos.x, newPos.z, PLAYER_RADIUS)) {
					newPos.y = uc->game->world->getFloorPos(curFloor);
					feetOnGround = true;
					jumpCount = 0;
					speed.y = 0;
				}
			}
		}
	}

	pos = newPos;
	// update orientation
	if (speed.length() > MINIMUM_SPEED) {
		float angle = rad2deg(acos(speed.x / speed.length()));
		if (speed.z > 0.f) {
			angle = -angle;
		}
		angle -= 90.f;

		if (fabs(angle - rot.y) > 180.f) {
			angle += 360.f * (angle > rot.y ? -1 : 1);
		}

		rot.y = smoothChange(rot.y, angle, PLAYER_ACCEL_ROT, uc->delta);
	}

	// reset controls
	wantJump = false;
	controlX = controlY = 0.f;

	Mesh::update(uc);
}


///////////////////////////////////////

Enemy::Enemy() : Guy("data/evilguy.msh") {
	maxSpeed = ENEMY_SPEED;
	type = CREATURE_ENEMY;
}

Enemy::~Enemy() {
}

void Enemy::update(UpdateContext *uc) {
	if (uc->gameMode == Game::GM_INIT)
		return;

	int curFloor = uc->game->world->getFloorNb(pos.y);
	int playerFloor = uc->game->world->getFloorNb(gGame->player->pos.y);

	Vector control;

	if (playerFloor == curFloor) {
		Vector targetAquired = gGame->player->pos - pos;

		if (gGame->mode == Game::GM_NORMAL && targetAquired.length() < DEADLY_RADIUS && fabs(targetAquired.y) < 80.f) {
			gGame->mode = Game::GM_DYING;
			gGame->player->speed = Vector(-500, 2500, 300);
			gGame->player->feetOnGround = false;
		}

		targetAquired.y = 0.f;
		if (targetAquired.length() > AI_RADIUS) {
			targetAquired.normalize();
			control.x = min(1.f, max(-1.f, targetAquired.x));
			control.y = min(1.f, max(-1.f, targetAquired.z));
		}
	}

	Map *map = uc->game->world->getMap(curFloor); 
	if (map) {
		Vector aiVision = pos + Vector(control.x, 0, control.y) * 50.f;
		int tileType = map->getTile(aiVision.x, aiVision.z, 10.f);
		if (tileType == 0) {
			Vector aiVision = pos + Vector(control.x, 0, 0) * 50.f;
			tileType = map->getTile(aiVision.x, aiVision.z, 10.f);
			if (tileType == 0) {
				control.x = 0.;
				if (fabs(gGame->player->pos.z - pos.z) < AI_RADIUS) {
					control.y = 0.;
				}
			}

			aiVision = pos + Vector(0, 0, control.y) * 50.f;
			tileType = map->getTile(aiVision.x, aiVision.z, 10.f);
			if (tileType == 0) {
				control.y = 0.;
				if (fabs(gGame->player->pos.x - pos.x) < AI_RADIUS) {
					control.x = 0.;
				}
			}
		}
	}

	//if (control.length() > 0.01) {
	//	control.normalize();
	//}

	controlX = control.x;
	controlY = control.y;

	Guy::update(uc);
}
